/* railroad crossing gate thingy, railroad.c */
/* dtmf 0 = gate up */
/* dtmf 1 = gate down */

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>
#include <sys/wait.h>
#include <sys/signal.h>

#ifdef __FreeBSD__
#include <machine/cpufunc.h>
#else
#include <asm/io.h>		/* GNU/Linux */
#endif

#define BASE_ADDRESS    0x378	/* base address of parallel port */
#define TONES           "/usr/local/bin/tones"	/* path to tones binary */

int c, status;
extern int errno;
			    /*       pin 432  on the parallel port */
unsigned char data_0 = 0;	/* Send 00000000b  up    */
unsigned char data_1 = 1;	/* Send 00000001b  down     */
char *tones = TONES;
pid_t pid;

int Up (void);
int Down (void);
int porttest (void);
void onalarm (int);
void usage (char *);

int
main (int argc, char **argv)
{

  if (geteuid () && getuid ())
    {
      fprintf (stderr,
	       "\n\n%s must be run as root to access hardware ports.\n",
	       argv[0]);
      fprintf (stderr, "Leaving...\n\n");
      exit (1);
    }

  if (argc < 2)
    {
      usage (argv[0]);
    }

  while ((c = getopt (argc, argv, "ud")) != EOF)
    {
      switch (c)
	{

	case 'u':
	  Up ();
	  exit (0);
	case 'd':
	  Down ();
	  exit (0);
	default:
	  usage (argv[0]);
	}
    }

  return (0);
}

int
Up (void)
{

  porttest ();

#ifdef __FreeBSD__
  outb (BASE_ADDRESS, data_0);
#else
  outb (data_0, BASE_ADDRESS);
  ioperm (BASE_ADDRESS, 3, 0);
#endif

  sleep (1);

  if ((pid = fork ()) < 0)
    perror ("fork");
  else if (pid == 0)
    {
      /* `tones sin :1000 941,1336` for dtmf 0 */
      if (execl (tones, tones, "sin", ":1000", "941,1336", (char *) 0) < 0)
	perror ("execl");
    }

  signal (SIGALRM, onalarm);

  alarm (2);

  if (wait (&status) == -1 || (status & 0177) != 0)
    fprintf (stderr, "%s killed.\n", tones);

  return (0);
}

int
Down (void)
{

  porttest ();

#ifdef __FreeBSD__
  outb (BASE_ADDRESS, data_1);
#else
  outb (data_1, BASE_ADDRESS);
  ioperm (BASE_ADDRESS, 3, 0);
#endif

  sleep (1);

  if ((pid = fork ()) < 0)
    perror ("fork");
  else if (pid == 0)
    {
      /* `tones sin :1000 697,1209` for dtmf 1 */
      if (execl (tones, tones, "sin", ":1000", "697,1209", (char *) 0) < 0)
	perror ("execl");
    }

  signal (SIGALRM, onalarm);

  alarm (2);

  if (wait (&status) == -1 || (status & 0177) != 0)
    fprintf (stderr, "%s killed.\n", tones);

  return (0);
}

int
porttest (void)
{

#ifdef __FreeBSD__
  if (open ("/dev/io", O_RDWR) < 0)
#else
  if (ioperm (BASE_ADDRESS, 3, 1) < 0)
#endif
    {
      fprintf (stderr, "\nERROR: Can not open port 0x%x\n\n", BASE_ADDRESS);
      perror ("port");
      exit (errno);
    }
  return (0);
}

void
usage (char *pname)
{
  fprintf (stderr, "Usage: %s\t [-u (gate up)] [-d (gate down)]\n", pname);
  exit (1);
}

void
onalarm (int val)
{
  kill (pid, SIGKILL);
}
